/********************************************************************************************************
 * @file     app.c
 *
 * @brief    for TLSR chips
 *
 * @author	 BLE Group
 * @date     May. 22, 2019
 *
 * @par      Copyright (c) Telink Semiconductor (Shanghai) Co., Ltd.
 *           All rights reserved.
 *
 *			 The information contained herein is confidential and proprietary property of Telink
 * 		     Semiconductor (Shanghai) Co., Ltd. and is available under the terms
 *			 of Commercial License Agreement between Telink Semiconductor (Shanghai)
 *			 Co., Ltd. and the licensee in separate contract or the terms described here-in.
 *           This heading MUST NOT be removed from this file.
 *
 * 			 Licensees are granted free, non-transferable use of the information in this
 *			 file under Mutual Non-Disclosure Agreement. NO WARRENTY of ANY KIND is provided.
 *
 *******************************************************************************************************/

#include "tl_common.h"
#include "drivers.h"
#include "stack/ble/ble.h"
#include "vendor/common/blt_common.h"
#include "battery_check.h"

#include "app_att.h"
#include "app_device.h"
#include "app_ui.h"
#include "custom_pair.h"


//all connections share same RX FIFO
MYFIFO_INIT(blt_rxfifo, 64, 16);


//different connections have their own TX FIFO, master and slave set independently
MULTI_CONN_FIFO_INIT(blt_m_txfifo, 40, 8, MASTER_MAX_NUM);  //master connection TX FIFO,  40*8*MASTER_MAX_NUM
MULTI_CONN_FIFO_INIT(blt_s_txfifo, 40, 8, SLAVE_MAX_NUM);   //slave  connection TX FIFO,  40*8*SLAVE_MAX_NUM


#define	MTU_M_BUFF_SIZE_MAX			ATT_ALLIGN4_DMA_BUFF(23)
#define MTU_S_BUFF_SIZE_MAX			ATT_ALLIGN4_DMA_BUFF(23) //if support LE Secure Connections, MTU must more than 65

_attribute_data_retention_	u8 mtu_m_rx_fifo[MASTER_MAX_NUM * MTU_M_BUFF_SIZE_MAX];

_attribute_data_retention_	u8 mtu_s_rx_fifo[SLAVE_MAX_NUM * MTU_S_BUFF_SIZE_MAX];
_attribute_data_retention_	u8 mtu_s_tx_fifo[SLAVE_MAX_NUM * MTU_S_BUFF_SIZE_MAX];


int	master_smp_pending = 0; 		// SMP: security & encryption;
int	master_sdp_pending = 0;			// SDP: service discovery


u32	lowBattDet_tick   = 0;




const u8	tbl_advData[] = {
	 0x09, 0x09, 'T','L','K','_','M','1','S','1',
	 0x02, 0x01, 0x05, 							// BLE limited discoverable mode and BR/EDR not supported
	 0x03, 0x19, 0x80, 0x01, 					// 384, Generic Remote Control, Generic category
	 0x05, 0x02, 0x12, 0x18, 0x0F, 0x18,		// incomplete list of service class UUIDs (0x1812, 0x180F)
};

const u8	tbl_scanRsp [] = {
	 0x09, 0x09, 'T','L','K','_','M','1','S','1'
};




const u8 	telink_adv_trigger_paring[] = {5, 0xFF, 0x11, 0x02, 0x01, 0x00};
const u8 	telink_adv_trigger_unpair[] = {5, 0xFF, 0x11, 0x02, 0x01, 0x01};

const u8 	telink_adv_trigger_paring_8258[] = {7, 0xFF, 0x11, 0x02, 0x01, 0x00, 0x58, 0x82};
const u8 	telink_adv_trigger_unpair_8258[] = {7, 0xFF, 0x11, 0x02, 0x01, 0x01, 0x58, 0x82};



int app_controller_event_callback (u32 h, u8 *p, int n);
int app_host_event_callback (u32 h, u8 *para, int n);
int app_gatt_data_handler (u16 connHandle, u8 *pkt);


int main_idle_loop (void);






int app_le_adv_report_event_handle(u8 *p)
{
	event_adv_report_t *pa = (event_adv_report_t *)p;
	s8 rssi = pa->data[pa->len];


	//// Here user can use ADV report data to do a lot of work ////



	/*********************** Master Create connection demo: Key press or ADV pair packet triggers pair  ********************/
	if(master_smp_pending || master_sdp_pending){ 	 //if previous connection SMP & SDP not finish, can not create a new connection
		return 1;
	}
	if (master_disconnect_connhandle){ //one master connection is in un_pair disconnection flow, do not create a new one
		return 1;
	}

	int master_auto_connect = 0;
	int user_manual_paring = 0;

	//manual paring methods 1: key press triggers
	user_manual_paring = master_pairing_enable;// && (rssi > -56);  //button trigger pairing(RSSI threshold, short distance)

	//manual paring methods 2: special paring ADV data
	if(!user_manual_paring){  //special ADV pair data can also trigger pairing
		user_manual_paring = (memcmp(pa->data, telink_adv_trigger_paring, sizeof(telink_adv_trigger_paring)) == 0) && (rssi > -56);
	}


	#if (BLE_MASTER_SMP_ENABLE)
		master_auto_connect = blc_smp_searchBondingSlaveDevice_by_PeerMacAddress(pa->adr_type, pa->mac);
	#else
		//search in slave mac_address table to find whether this device is an old device which has already paired with master
		master_auto_connect = user_tbl_slave_mac_search(pa->adr_type, pa->mac);
	#endif


	if(master_auto_connect || user_manual_paring){

		/* send create connection command to Controller, trigger it switch to initiating state. After this command, Controller
		 * will scan all the ADV packets it received but not report to host, to find the specified device(mac_adr_type & mac_adr),
		 * then send a "CONN_REQ" packet, enter to connection state and send a connection complete event
		 * (HCI_SUB_EVT_LE_CONNECTION_COMPLETE) to Host*/
		u8 status = blc_llms_createConnection( SCAN_INTERVAL_100MS, SCAN_INTERVAL_100MS, INITIATE_FP_ADV_SPECIFY,  \
								 pa->adr_type, pa->mac, BLE_ADDR_PUBLIC, \
								 CONN_INTERVAL_11P25MS, CONN_INTERVAL_11P25MS, 0, CONN_TIMEOUT_4S, \
								 0, 0xFFFF);


		if(status == BLE_SUCCESS){ //create connection success
			#if (!BLE_MASTER_SMP_ENABLE)
			    // for Telink referenced pair&bonding,
				if(user_manual_paring && !master_auto_connect){  //manual pair but not auto connect
					blm_manPair.manual_pair = 1;
					blm_manPair.mac_type = pa->adr_type;
					memcpy(blm_manPair.mac, pa->mac, 6);
					blm_manPair.pair_tick = clock_time();
				}
			#endif
		}
	}
	/*********************** Master Create connection demo code end  *******************************************************/


	return 0;
}



int app_le_connection_complete_event_handle(u8 *p)
{
	event_connection_complete_t *pCon = (event_connection_complete_t *)p;

	if(pCon->status == BLE_SUCCESS){

		dev_char_info_t cur_conn_device;
		memset(&cur_conn_device, 0, sizeof(dev_char_info_t));

		//save current connect address type and address and conn_handle
		cur_conn_device.conn_handle = pCon->handle;
		cur_conn_device.peer_adrType = pCon->peer_adr_type;
		memcpy(cur_conn_device.peer_addr, pCon->mac, 6);


		dev_char_info_insert( &cur_conn_device );


		if( pCon->handle & BLM_CONN_HANDLE ) // master connection handle, process SMP and SDP
		{
			memset(&cur_master_conn_device, 0, sizeof(dev_char_info_t));

			//save current connect address type and address and conn_handle
			cur_master_conn_device.conn_handle = pCon->handle;
			cur_master_conn_device.peer_adrType = pCon->peer_adr_type;
			memcpy(cur_master_conn_device.peer_addr, pCon->mac, 6);


			#if (BLE_MASTER_SMP_ENABLE)
				master_smp_pending = pCon->handle; // this connection need SMP
			#else
				//manual paring, device match, add this device to slave mac table
				if(blm_manPair.manual_pair && blm_manPair.mac_type == pCon->peer_adr_type && !memcmp(blm_manPair.mac, pCon->mac, 6)){
					blm_manPair.manual_pair = 0;

					user_tbl_slave_mac_add(pCon->peer_adr_type, pCon->mac);
				}
			#endif



			#if (BLE_MASTER_SIMPLE_SDP_ENABLE)
				u8	temp_buff[sizeof(dev_att_t)];
				dev_att_t *pdev_att = (dev_att_t *)temp_buff;

				if( dev_char_info_search_peer_att_handle_by_peer_mac(pCon->peer_adr_type, pCon->mac, pdev_att) ){  //att_handle search in flash success, load char_handle directly from flash, no need SDP again
					//cur_master_conn_device.char_handle[0] = 									//MIC
					//cur_master_conn_device.char_handle[1] = 									//Speaker
					cur_master_conn_device.char_handle[2] = pdev_att->char_handle[2];			//OTA
					cur_master_conn_device.char_handle[3] = pdev_att->char_handle[3];			//consume report
					cur_master_conn_device.char_handle[4] = pdev_att->char_handle[4];			//normal key report
					//cur_master_conn_device.char_handle[6] =									//BLE Module, SPP Server to Client
					//cur_master_conn_device.char_handle[7] =									//BLE Module, SPP Client to Server

					dev_char_info_add_peer_att_handle(&cur_master_conn_device);  //add the peer device att_handle value to conn_dev_list
				}
				else{
					master_sdp_pending = pCon->handle;  // mark this connection need SDP

					#if (BLE_MASTER_SMP_ENABLE)
						 //service discovery initiated after SMP done, trigger it in "GAP_EVT_MASK_SMP_SECURITY_PROCESS_DONE" event callBack.
					#else
					     //No SMP, service discovery can initiated now
						 app_register_service(&app_service_discovery);
					#endif
				}

			#endif
		}
	}

	return 0;
}





int 	app_disconnect_event_handle(u8 *p)
{
	event_disconnection_t	*pd = (event_disconnection_t *)p;

	//terminate reason
	if(pd->reason == HCI_ERR_CONN_TIMEOUT){  	//connection timeout

	}
	else if(pd->reason == HCI_ERR_REMOTE_USER_TERM_CONN){  	//peer device send terminate command on link layer

	}
	//master host disconnect( blm_ll_disconnect(current_connHandle, HCI_ERR_REMOTE_USER_TERM_CONN) )
	else if(pd->reason == HCI_ERR_CONN_TERM_BY_LOCAL_HOST){

	}
	else{

	}


	//if previous connection SMP & SDP not finished, clear flag
	if(master_smp_pending == pd->connHandle){
		master_smp_pending = 0;
	}
	if(master_sdp_pending == pd->connHandle){
		master_sdp_pending = 0;
	}

	if(master_disconnect_connhandle == pd->connHandle){  //un_pair disconnection flow finish, clear flag
		master_disconnect_connhandle = 0;
	}

	dev_char_info_delete_by_connhandle(pd->connHandle);


	return 0;
}



//////////////////////////////////////////////////////////
// event call back
//////////////////////////////////////////////////////////
int app_controller_event_callback (u32 h, u8 *p, int n)
{
	if (h &HCI_FLAG_EVENT_BT_STD)		//Controller HCI event
	{
		u8 evtCode = h & 0xff;

		//------------ disconnect -------------------------------------
		if(evtCode == HCI_EVT_DISCONNECTION_COMPLETE)  //connection terminate
		{
			app_disconnect_event_handle(p);
		}
		else if(evtCode == HCI_EVT_LE_META)  //LE Event
		{
			u8 subEvt_code = p[0];

			//------hci le event: le connection complete event---------------------------------
			if (subEvt_code == HCI_SUB_EVT_LE_CONNECTION_COMPLETE)	// connection complete
			{
				//after controller is set to initiating state by host (blc_llms_createConnection(...) )
				//it will scan the specified device(adr_type & mac), when find this adv packet, send a connection request packet to slave
				//and enter to connection state, send connection complete event. but notice that connection complete not
				//equals to connection establish. connection complete means that master controller set all the ble timing
				//get ready, but has not received any slave packet, if slave rf lost the connection request packet, it will
				//not send any packet to master controller

				app_le_connection_complete_event_handle(p);

			}
			//------hci le event: le connection establish event---------------------------------
			else if(subEvt_code == HCI_SUB_EVT_LE_CONNECTION_ESTABLISH)  //connection establish(telink private event)
			{
				//notice that: this connection event is defined by telink, not a standard ble controller event
				//after master controller send connection request packet to slave, when slave received this packet
				//and enter to connection state, send a ack packet within 6 connection event, master will send
				//connection establish event to host(HCI_SUB_EVT_LE_CONNECTION_ESTABLISH)

				//	blc_le_connection_establish_event_handle(p);
			}
			//--------hci le event: le adv report event ----------------------------------------
			else if (subEvt_code == HCI_SUB_EVT_LE_ADVERTISING_REPORT)	// ADV packet
			{
				//after controller is set to scan state, it will report all the adv packet it received by this event

				app_le_adv_report_event_handle(p);
			}
			//------hci le event: le connection update complete event-------------------------------
			else if (subEvt_code == HCI_SUB_EVT_LE_CONNECTION_UPDATE_COMPLETE)	// connection update
			{

			}
		}
	}


	return 0;

}





int app_host_event_callback (u32 h, u8 *para, int n)
{
	u8 event = h & 0xFF;

	switch(event)
	{
		case GAP_EVT_SMP_PARING_BEAGIN:
		{
			gap_smp_paringBeginEvt_t *p = (gap_smp_paringBeginEvt_t *)para;
			/*
			 * 	JustWorks = 0, PK_Init_Dsply_Resp_Input = 1, PK_Resp_Dsply_Init_Input = 2, PK_BOTH_INPUT = 3,OOB_Authentication = 4, Numric_Comparison = 5,
			 */
			u8 TK_method[6][40] = {"JustWorks", "PK_Init_Dsply_Resp_Input", "PK_Resp_Dsply_Init_Input", "PK_BOTH_INPUT", "OOB_Authentication", "Numric_Comparison"};
			printf("[0x%x]Pairing begin (sc:%s,TK-method:%s)\n", p->connHandle, p->secure_conn?"SC":"Legacy", TK_method[p->tk_method]);
			printf("[0x%x]Pairing begin\n", p->connHandle);
		}
		break;

		case GAP_EVT_SMP_PARING_SUCCESS:
		{
			gap_smp_paringSuccessEvt_t* p = (gap_smp_paringSuccessEvt_t*)para;
		    printf("[0x%x]Pairing success: bond%s (save %s)\n", p->connHandle, p->bonding ?"T":"F", p->bonding_result ? "OK" : "ERR");
		}
		break;

		case GAP_EVT_SMP_PARING_FAIL:
		{
			gap_smp_paringFailEvt_t* p = (gap_smp_paringFailEvt_t*)para;
			printf("[0x%x]Pairing failed:rsn:0x%x\n", p->connHandle, p->reason);
		}
		break;

		case GAP_EVT_SMP_CONN_ENCRYPTION_DONE:
		{

		}
		break;


		case GAP_EVT_SMP_SECURITY_PROCESS_DONE:
		{
			gap_smp_connEncDoneEvt_t* p = (gap_smp_connEncDoneEvt_t*)para;
			//printf("[0x%x]ENC done (%s)\n", p->connHandle, p->re_connect == SMP_STANDARD_PAIR ? "1stParing" :"reConnect");

			if( p->connHandle & BLM_CONN_HANDLE){  //master connection
				#if (BLE_MASTER_SMP_ENABLE)
					if(master_smp_pending == p->connHandle){
						master_smp_pending = 0;
					}
				#endif

				#if (BLE_MASTER_SIMPLE_SDP_ENABLE)  //SMP finish
					if(master_sdp_pending == p->connHandle){  //SDP is pending
						app_register_service(&app_service_discovery);  //start SDP now
					}
				#endif
			}

		}
		break;



		case GAP_EVT_SMP_TK_DISPALY:
		{
			char pc[7];
			gap_smp_TkDisplayEvt_t* p = (gap_smp_TkDisplayEvt_t*)para;
			sprintf(pc, "%d", p->tk_pincode);
			//printf("[0x%x]TK display:%s\n", p->connHandle, pc);
		}
		break;

		case GAP_EVT_SMP_TK_REQUEST_PASSKEY:
		{
			gap_smp_TkReqPassKeyEvt_t* p = (gap_smp_TkReqPassKeyEvt_t*)para;
			//printf("[0x%x]TK Request passkey\n", p->connHandle);
		}
		break;

		case GAP_EVT_SMP_TK_REQUEST_OOB:
		{

		}
		break;

		case GAP_EVT_SMP_TK_NUMERIC_COMPARE:
		{

		}
		break;

		case GAP_EVT_ATT_EXCHANGE_MTU:
		{

		}
		break;

		case GAP_EVT_GATT_HANDLE_VLAUE_CONFIRM:
		{

		}
		break;

		default:
		break;
	}

	return 0;
}





int app_gatt_data_handler (u16 connHandle, u8 *pkt)
{

	if( connHandle & BLM_CONN_HANDLE)   //GATT data for Master
	{
			#if (BLE_MASTER_SIMPLE_SDP_ENABLE)
				if(master_sdp_pending == connHandle ){  //ATT service discovery is ongoing on this conn_handle
					//when service discovery function is running, all the ATT data from slave
					//will be processed by it,  user can only send your own att cmd after  service discovery is over
					host_att_client_handler (connHandle, pkt); //handle this ATT data by service discovery process
				}
			#endif

			//so any ATT data before service discovery will be dropped
			dev_char_info_t* dev_info = dev_char_info_search_by_connhandle (connHandle);
			if(dev_info)
			{
				//-------	user process ------------------------------------------------
				rf_packet_att_t *pAtt = (rf_packet_att_t*)pkt;
				u16 attHandle = pAtt->handle0 | pAtt->handle1<<8;

				if(pAtt->opcode == ATT_OP_HANDLE_VALUE_NOTI)  //slave handle notify
				{
						//---------------	consumer key --------------------------
					#if (BLE_MASTER_SIMPLE_SDP_ENABLE)
						if(attHandle == dev_info->char_handle[3])  // Consume Report In (Media Key)
					#else
						if(attHandle == 25)   //Demo device(825x_ble_sample) Consume Report AttHandle value is 25
					#endif
						{
							att_keyboard_media (connHandle, pAtt->dat);
						}
						//---------------	keyboard key --------------------------
					#if (BLE_MASTER_SIMPLE_SDP_ENABLE)
						else if(attHandle == dev_info->char_handle[4])     // Key Report In
					#else
						else if(attHandle == HID_HANDLE_KEYBOARD_REPORT)   // Demo device(825x_ble_sample) Key Report AttHandle value is 29
					#endif
						{
							att_keyboard (connHandle, pAtt->dat);
						}
						else
						{

						}
				}
				else if (pAtt->opcode == ATT_OP_HANDLE_VALUE_IND)
				{

				}
			}

	}
	else{   //GATT data for Slave


	}


	return 0;
}










void  app_set_sleep_wakeup (u8 e, u8 *p, int n)
{
	//suspend time too big, add GPIO wake_up to prevent some keyScan loss.
	if( ((u32)(blc_pm_getSystemWakeupTick() - clock_time())) > 50 * CLOCK_16M_SYS_TIMER_CLK_1MS){
		blc_pm_setWakeupSource(PM_WAKEUP_PAD);  //GPIO pad wake_up suspend
	}
}


extern u32	scan_pin_need;




///////////////////////////////////////////
void user_init(void)
{
	//random number generator must be initiated here( in the beginning of user_init_nromal)
	//when deepSleep retention wakeUp, no need initialize again
	random_generator_init();  //this is must

	/*****************************************************************************************
	 Note: battery check must do before any flash write/erase operation, cause flash write/erase
		   under a low or unstable power supply will lead to error flash operation

		   Some module initialization may involve flash write/erase, include: OTA initialization,
				SMP initialization, ..
				So these initialization must be done after  battery check
	*****************************************************************************************/
	#if (BATT_CHECK_ENABLE)  //battery check must do before OTA relative operation
		if(analog_read(USED_DEEP_ANA_REG) & LOW_BATT_FLG){
			app_battery_power_check(VBAT_ALRAM_THRES_MV + 200);  //2.2 V
		}
		else{
			app_battery_power_check(VBAT_ALRAM_THRES_MV);  //2.0 V
		}
	#endif


///////////// BLE stack Initialization ////////////////
	//for 512K Flash, flash_sector_mac_address equals to 0x76000
	//for 1M   Flash, flash_sector_mac_address equals to 0xFF000
	u8  mac_public[6];
	u8  mac_random_static[6];
	blc_initMacAddress(flash_sector_mac_address, mac_public, mac_random_static);


	////// Controller Initialization  //////////
	blc_llms_initBasicMCU();
	blc_llms_initStandby_module(mac_public);						   //mandatory

	blc_llms_initAdvertising_module(); 	//adv module: 		 mandatory for BLE slave,

	blc_llms_initScanning_module(); 		//scan module: 		 mandatory for BLE master,

	blc_llms_initMultiMasterMultiSlave_module();

	blc_llms_setMaxConnectionNumber( MASTER_MAX_NUM, SLAVE_MAX_NUM);



	rf_set_power_level_index (RF_POWER_P3p01dBm);


	////// HCI Initialization  //////////
	blc_hci_registerControllerDataHandler (blt_l2cap_pktHandler);

	blc_hci_registerControllerEventHandler(app_controller_event_callback); //controller hci event to host all processed in this func

	//bluetooth event
	blc_hci_setEventMask_cmd (HCI_EVT_MASK_DISCONNECTION_COMPLETE);

	//bluetooth low energy(LE) event
	blc_hci_le_setEventMask_cmd(		HCI_LE_EVT_MASK_CONNECTION_COMPLETE  \
									|	HCI_LE_EVT_MASK_ADVERTISING_REPORT \
									|   HCI_LE_EVT_MASK_CONNECTION_UPDATE_COMPLETE \
									|   HCI_LE_EVT_MASK_CONNECTION_ESTABLISH);  //connection establish: telink private event   TODO: delete this cause we will support standard HCI







	////// Host Initialization  //////////
	blc_gap_init();    //GAP initialization must be done before any other host feature initialization !!!

	//l2cap initialization
	blm_l2cap_initMtuBuffer(mtu_m_rx_fifo, MTU_M_BUFF_SIZE_MAX, 0,0);
	bls_l2cap_initMtuBuffer(mtu_s_rx_fifo, MTU_S_BUFF_SIZE_MAX, mtu_s_tx_fifo, MTU_S_BUFF_SIZE_MAX);
	my_gatt_init();  //gatt initialization
	#if (BLE_MASTER_SIMPLE_SDP_ENABLE)
		host_att_register_idle_func (main_idle_loop);
	#endif
	blc_gatt_register_data_handler(app_gatt_data_handler);


	//SMP initialization
	#if (BLE_SLAVE_SMP_ENABLE)  //Slave SMP Enable
		blc_smp_setSecurityLevel_slave(Unauthenticated_Paring_with_Encryption);  //LE_Security_Mode_1_Level_2
	#else
		blc_smp_setSecurityLevel_slave(No_Security);
	#endif

	#if (BLE_MASTER_SMP_ENABLE) //Master SMP Enable
		blc_smp_setSecurityLevel_master(Unauthenticated_Paring_with_Encryption);  //LE_Security_Mode_1_Level_2
	#else
		blc_smp_setSecurityLevel_master(No_Security);
		user_master_host_pairing_management_init(); 		//TeLink referenced paring&bonding without standard paring in BLE Spec
	#endif

	blc_smp_smpParamInit();


	//host(GAP/SMP/GATT/ATT) event process: register host event callback and set event mask
	blc_gap_registerHostEventHandler( app_host_event_callback );
	blc_gap_setEventMask( GAP_EVT_MASK_SMP_PARING_BEAGIN 			|  \
						  GAP_EVT_MASK_SMP_PARING_SUCCESS   		|  \
						  GAP_EVT_MASK_SMP_PARING_FAIL				|  \
						  GAP_EVT_MASK_SMP_SECURITY_PROCESS_DONE);



	////// ADV and Scan CONFIG  //////////
	blc_llms_setAdvData( (u8 *)tbl_advData, sizeof(tbl_advData) );
	blc_llms_setScanRspData( (u8 *)tbl_scanRsp, sizeof(tbl_scanRsp));



	blc_llms_setAdvParam(  ADV_INTERVAL_30MS, ADV_INTERVAL_30MS,
											 ADV_TYPE_CONNECTABLE_UNDIRECTED, OWN_ADDRESS_PUBLIC,
											 0,  NULL,
											 BLT_ENABLE_ADV_ALL,
											 ADV_FP_NONE);
	blc_llms_setAdvEnable(1);  //ADV enable


	blc_llms_setScanParameter(SCAN_TYPE_PASSIVE, SCAN_INTERVAL_200MS, SCAN_WINDOW_50MS, OWN_ADDRESS_PUBLIC, SCAN_FP_ALLOW_ADV_ANY);
	blc_llms_setScanEnable (BLC_SCAN_ENABLE, DUP_FILTER_DISABLE);




#if (BLE_APP_PM_ENABLE)
	blc_llms_initPowerManagement_module();

	blc_llms_registerTelinkControllerEventCallback (BLT_EV_FLAG_SUSPEND_ENTER, &app_set_sleep_wakeup);
#else
	blc_pm_setSuspendMask (PM_SUSPEND_DISABLE);
#endif


	app_ui_init();
}



/////////////////////////////////////////////////////////////////////
// main loop flow
/////////////////////////////////////////////////////////////////////
int main_idle_loop (void)
{
	////////////////////////////////////// BLE entry /////////////////////////////////
	blt_llms_sdk_main_loop();

	////////////////////////////////////// UI entry /////////////////////////////////
	#if (UI_KEYBOARD_ENABLE)
		proc_keyboard (0,0, 0);
	#endif

	proc_master_role_unpair();

	#if (BATT_CHECK_ENABLE)
		if(battery_get_detect_enable() && clock_time_exceed(lowBattDet_tick, 500000) ){
			lowBattDet_tick = clock_time();
			app_battery_power_check(VBAT_ALRAM_THRES_MV);  //2000 mV low battery
		}
	#endif



	#if(BLE_MASTER_OTA_ENABLE)

	#endif


	////////////////////////////////////// PM Process /////////////////////////////////
//	blc_pm_setSuspendMask( PM_SUSPEND_ADV );
//	blc_pm_setSuspendMask( PM_SUSPEND_SCAN );
	blc_pm_setSuspendMask( PM_SUSPEND_ADV | PM_SUSPEND_SCAN | PM_SUSPEND_SLAVE | PM_SUSPEND_MASTER);


	return 0;
}




void main_loop (void)
{
	main_idle_loop ();

	if (main_service)
	{
		main_service ();
		main_service = 0;
	}
}





