/********************************************************************************************************
 * @file	 app_device.c
 *
 * @brief    for TLSR chips
 *
 * @author	 BLE Group
 * @date     May. 22, 2019
 *
 * @par      Copyright (c) Telink Semiconductor (Shanghai) Co., Ltd.
 *           All rights reserved.
 *
 *			 The information contained herein is confidential and proprietary property of Telink
 * 		     Semiconductor (Shanghai) Co., Ltd. and is available under the terms
 *			 of Commercial License Agreement between Telink Semiconductor (Shanghai)
 *			 Co., Ltd. and the licensee in separate contract or the terms described here-in.
 *           This heading MUST NOT be removed from this file.
 *
 * 			 Licensees are granted free, non-transferable use of the information in this
 *			 file under Mutual Non-Disclosure Agreement. NO WARRENTY of ANY KIND is provided.
 *
 *******************************************************************************************************/

#include "tl_common.h"
#include "drivers.h"
#include "stack/ble/ble.h"


#include "app_att.h"
#include "app_device.h"







/*
 * Used for store information of connected devices.
 *
 * 0 ~ (MASTER_MAX_NUM - 1) is for master,  MASTER_MAX_NUM ~ (MASTER_MAX_NUM + SLAVE_MAX_NUM - 1) s for slave
 *
 * e.g.    MASTER_MAX_NUM	SLAVE_MAX_NUM			master					  slave
 *                1				  1				conn_dev_list[0]		conn_dev_list[1]
 *                2				  2				conn_dev_list[0..1]		conn_dev_list[2..3]
 *                3				  2				conn_dev_list[0..2]		conn_dev_list[3..4]
 *                4				  3				conn_dev_list[0..3]		conn_dev_list[4..6]
 */
dev_char_info_t	conn_dev_list[DEVICE_CHAR_INFO_MAX_NUM];

/*
 * Special Design: when a device connection terminate, e.g. device index is 1, we will not only remove conn_dev_list[1],
 * 				   but also copy conn_dev_list[2] (assume that this device is in connection state) to conn_dev_list[1],
 * 				   			copy conn_dev_list[3] (assume that this device is in connection state) to conn_dev_list[2],
 * 				   			......
 *
 * So for 4 master 3 slave demo, the relationship between device in connection state and conn_master_num/conn_slave_num is as follows
 *
 * 			conn_master_num	   					device in connection state
 *
 * 				   1		  						conn_dev_list[0]
 * 				   2								conn_dev_list[0]/conn_dev_list[1]
 * 				   3								conn_dev_list[0]/conn_dev_list[1]/conn_dev_list[2]
 * 				   4								conn_dev_list[0]/conn_dev_list[1]/conn_dev_list[2]/conn_dev_list[3]
 *
 * 			conn_slave_num
 *
 * 				   1		  						conn_dev_list[4]
 * 				   2								conn_dev_list[4]/conn_dev_list[5]
 * 				   3								conn_dev_list[4]/conn_dev_list[5]/conn_dev_list[6]
 *
 *     			.conn_state must be 1, conn_dev_list[2].conn_state
 */
int conn_master_num = 0;   //current master number in connection state
int conn_slave_num = 0;    //current slave number in connection state




dev_char_info_t cur_conn_device;




/*
 * Used for add device information to conn_dev_list.
 */
int dev_char_info_insert (dev_char_info_t* dev_char_info)
{
	int index = 100;
	if( dev_char_info->conn_handle & BLM_CONN_HANDLE ){  //master
		if(conn_master_num < MASTER_MAX_NUM){
			index = conn_master_num;
			conn_master_num ++;
		}
	}
	else if( dev_char_info->conn_handle & BLS_CONN_HANDLE ){  //slave
		if(conn_slave_num < SLAVE_MAX_NUM){
			index = MASTER_MAX_NUM + conn_slave_num;
			conn_slave_num ++;
		}
	}

	if(index < MASTER_MAX_NUM + SLAVE_MAX_NUM){
		memcpy(&conn_dev_list[index], dev_char_info, sizeof(dev_char_info_t));
		conn_dev_list[index].conn_state = 1;

		return 0;   //success
	}
	else{
		return 1;   //fail
	}

}





/*
 * Used for add peer device service ATThandle.
 */
int dev_char_info_add_peer_att_handle (dev_char_info_t* dev_char_info)
{
	int i;
	for(i=0; i< conn_master_num; i++){
		if( conn_dev_list[i].conn_handle == dev_char_info->conn_handle){
			break;
		}
	}

	if( i < conn_master_num){
		for(int j=0; j<CHAR_HANDLE_MAX;  j++){
			conn_dev_list[i].char_handle[j] = dev_char_info->char_handle[j];
		}

		conn_dev_list[i].char_handle_valid = 1;

		return 0;   //success
	}
	else{
		return 1;   //fail
	}
}




/*
 * Used for delete device information from conn_dev_list by connHandle
 */
int dev_char_info_delete_by_connhandle (u16 connhandle)
{
	foreach (i, DEVICE_CHAR_INFO_MAX_NUM)
	{
		if(conn_dev_list[i].conn_handle == connhandle && conn_dev_list[i].conn_state)  //match
		{
//			memset(&conn_dev_list[i], 0, sizeof(dev_char_info_t));

			if(connhandle & BLM_CONN_HANDLE){   //master
				memcpy (&conn_dev_list[i], &conn_dev_list[i + 1], (conn_master_num - i - 1) * sizeof (dev_char_info_t));
				conn_master_num --;
				memset(&conn_dev_list[conn_master_num], 0, sizeof(dev_char_info_t));
			}
			else{  //slave
				memcpy (&conn_dev_list[i], &conn_dev_list[i + 1], (conn_slave_num + MASTER_MAX_NUM - i - 1) * sizeof (dev_char_info_t));
				conn_slave_num --;
				memset(&conn_dev_list[conn_slave_num + MASTER_MAX_NUM], 0, sizeof(dev_char_info_t));
			}

			return 0;
		}
	}
	return 1;  //no find the peer device address.
}


/*
 * Used for delete device information from conn_dev_list by peer device's mac address
 */
int 	dev_char_info_delete_by_peer_mac_address (u8 adr_type, u8* addr)
{
	foreach (i, DEVICE_CHAR_INFO_MAX_NUM)
	{
		if(conn_dev_list[i].conn_state)
		{
			int mac_match = 0;
			if( IS_RESOLVABLE_PRIVATE_ADDR(adr_type, addr) ){
				//TODO
			}
			else{
				if(adr_type == conn_dev_list[i].peer_adrType && (!memcmp (addr, conn_dev_list[i].peer_addr, 6)) ){
					mac_match = 1;
				}
			}

			u16 connhandle = conn_dev_list[i].conn_handle;

			if(mac_match){
				if(connhandle & BLM_CONN_HANDLE){   //master
					memcpy (&conn_dev_list[i], &conn_dev_list[i + 1], (conn_master_num - i - 1) * sizeof (dev_char_info_t));
					memset(&conn_dev_list[conn_master_num], 0, sizeof(dev_char_info_t));
					conn_master_num --;
				}
				else{  //slave
					memcpy (&conn_dev_list[i], &conn_dev_list[i + 1], (conn_slave_num + MASTER_MAX_NUM - i - 1) * sizeof (dev_char_info_t));
					memset(&conn_dev_list[conn_slave_num + MASTER_MAX_NUM], 0, sizeof(dev_char_info_t));
					conn_slave_num --;
				}

				return 0;
			}

		}
	}

	return 1;  //no find the peer device address.
}







/*
 * Used for search device information from conn_dev_list by connHandle.
 * return  - 0 no peer device.
 * 			 others - device info
 */
dev_char_info_t* dev_char_info_search_by_connhandle (u16 connhandle)
{
	foreach (i, DEVICE_CHAR_INFO_MAX_NUM)
	{
		if(conn_dev_list[i].conn_handle == connhandle && conn_dev_list[i].conn_state)
		{
			return &conn_dev_list[i];   // find the peer device
		}
	}

	return 0;  // no find the peer device
}




/*
 * Used for search device information from conn_dev_list by peer device mac_address
 */
dev_char_info_t* dev_char_info_search_by_peer_mac_address (u8 adr_type, u8* addr)
{
	foreach (i, DEVICE_CHAR_INFO_MAX_NUM)
	{

		int mac_match = 0;
		if( IS_RESOLVABLE_PRIVATE_ADDR(adr_type, addr) ){
			//TODO
		}
		else{
			if(adr_type == conn_dev_list[i].peer_adrType && (!memcmp (addr, conn_dev_list[i].peer_addr, 6)) ){
				mac_match = 1;
			}
		}

		if(mac_match){
			return &conn_dev_list[i];   // find the peer device
		}
	}

	return 0;  // no find the peer device
}








